/*
 * MPS-Pico8 Linux Driver
 *
 *  Copyright (C) 2018 CAEN ELS s.r.l.
 *
 *  Paolo Scarbolo <p.scarbolo@caenels.com>
 *
 *  Copyright 2016 Board of Trustees of Michigan State University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License v2 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "mps_pico_char.h"

static int char_open (
    struct inode *inode,
    struct file *file
)
{
    struct file_data *fdata;
	
    struct board_data *board = container_of(inode->i_cdev, struct board_data, cdev);

	dev_info(&board->pci_dev->dev, "char_open()\n");

	if (!try_module_get(THIS_MODULE))
    {
		return -ENODEV;
    }

    fdata = kzalloc(sizeof(*fdata), GFP_KERNEL);

    if (!fdata) 
    {
        module_put(THIS_MODULE);
        return -ENOMEM;
    }
    
    fdata->board = board;

    file->private_data = fdata;

	return 0;
}

static int char_release (
    struct inode *inode, 
    struct file *file
)
{
    struct file_data *fdata = (struct file_data *)file->private_data;
	struct board_data *board = container_of(inode->i_cdev, struct board_data, cdev);

	dev_info(&board->pci_dev->dev, "char_release()\n");

    kfree(fdata);

	module_put(THIS_MODULE);
	return 0;
}

#ifdef CONFIG_MPS_PICO_FRIB
static ssize_t frib_write_reg(struct board_data *board,
                             const char __user *buf,
                             size_t count,
                             loff_t *pos);
static ssize_t frib_read_reg(struct board_data *board,
                             char __user *buf,
                             size_t count,
                             loff_t *pos);
static ssize_t frib_read_capture(struct board_data *board,
                                 char __user *buf,
                                 size_t count,
                                 loff_t *pos);
#endif




////////////////////////////////////////////////
// Read from DMA1 and DMA2
////////////////////////////////////////////////
static ssize_t char_read (
	struct file *filp,
	char __user *buf,
	size_t count,
	loff_t *pos
)
{
    struct file_data *fdata = (struct file_data *)filp->private_data;
    struct board_data *board = fdata->board;
	int rc, cond1, cond2;
	size_t tmp_count, dma1_count, dma2_count, dma1_count_copy;
    const int ratio = 1000;
	int i;

    dev_dbg(&board->pci_dev->dev, "  read(), site_mode=%u count %zd\n", fdata->site_mode, count);
    if ( fdata->site_mode != 0 )
    {
        return -EINVAL;
    }

    // if read bytes larger than MAX, exit
    if (count > DMA_BUF_COUNT*DMA_BUF_SIZE) 
    {
        return -EINVAL;
    }

    // check if DMA1 read in progress
    spin_lock_irq(&board->dma_queue.lock);
	if (board->read1_in_progress) 
    {
        spin_unlock_irq(&board->dma_queue.lock);
        dev_warn(&board->pci_dev->dev, "  DMA1 read(), concurrent read()s not allowed\n");
		return -EIO;
	}
	board->read1_in_progress = 1;
    // check if DMA2 read in progress
    if (board->read2_in_progress) 
    {
        spin_unlock_irq(&board->dma_queue.lock);
        dev_warn(&board->pci_dev->dev, "  DMA2 read(), concurrent read()s not allowed\n");
		return -EIO;
	}
	board->read2_in_progress = 1;

	// PREPARE DMA1 FOR TRANSFER
	i = 0;
	tmp_count = count;
    dma2_count = ((tmp_count / ratio) / DMA_BYTES_PER_LINE) * DMA_BYTES_PER_LINE;
    dma1_count = tmp_count - dma2_count;
    dma1_count_copy = dma1_count;
    // disable DMA engine
    dma1_enable(board, 0);
    dma2_enable(board, 0);
	while (dma1_count > DMA_BUF_SIZE) 
    {
        dma1_push(board, (uint32_t)board->dma1_buf[i++], DMA_BUF_SIZE, 0);
		dma1_count -= DMA_BUF_SIZE;
	}
    // write ADDRESS, LENGTH AND RISE IRQ
	dma1_push(board, (uint32_t)board->dma1_buf[i], dma1_count, 1);
    dma2_push(board, (uint32_t)board->dma2_buf, dma2_count, 1);
    mb();
    dev_info(&board->pci_dev->dev, "  dma1_push(): length DMA1/2 transfer %lu %lu\n", dma1_count, dma2_count);
    // ENABLE DMA engine
	dma1_enable(board, 1);
    dma2_enable(board, 1);
    mb();


    //////////////////////////////////////////////////////////
    // check DMA1
    //////////////////////////////////////////////////////////
    if (likely(board->irqmode != dmac_irq_poll)) 
    {
        //const unsigned long twait2 = msecs_to_jiffies(2);
        // TODO: For the moment I check only the dma1_queue, check also dma2_queue??
        // wait_event_interruptible_locked_irq - sleep until a condition gets true 
        // The function will return -ERESTARTSYS if it was interrupted by a signal and 0 if condition evaluated to true.
        dev_dbg(&board->pci_dev->dev, "  start wait_event_interruptible_locked_irq on DMA1\n");
        rc = wait_event_interruptible_locked_irq(board->dma_queue, board->dma1_irq_flag!=0);
        dev_dbg(&board->pci_dev->dev, "  wait_event_interruptible_locked_irq on DMA1 rc = %d\n", rc);
    } 
    else 
    {
        const unsigned long twait1 = msecs_to_jiffies(1);
        do 
        {
            spin_unlock_irq(&board->dma_queue.lock);
            /* must unlock for call to amc_isr() as spin locks aren't recursive */
            if (amc_isr(board->pci_dev->irq, board) == IRQ_NONE)
            {
                // wait_event_interruptible_timeout - sleep until a condition gets true or a timeout elapses
                // 0 if the condition evaluated to false after the timeout elapsed, 1 if the condition evaluated 
                // to true after the timeout elapsed, the remaining jiffies (at least 1) if the condition evaluated 
                // to true before the timeout elapsed, or -ERESTARTSYS if it was interrupted by a signal
                rc = wait_event_interruptible_timeout(board->dma_queue, board->dma1_irq_flag!=0, twait1);
                dev_dbg(&board->pci_dev->dev, "  wait_event_interruptible_timeout on DMA1 rc = %d\n", rc);
            }
            else
            {
                rc = 0;
            }
            spin_lock_irq(&board->dma_queue.lock);
            /* continue while no "IRQ" signaled, and wait not interrupted */
        } while(board->dma1_irq_flag==0 && rc>=0);

        if (rc > 0)
        {
            rc = 0;
        }

    }

    // re-save the initial value of dma1_count
    dma1_count = dma1_count_copy;
    /*
     * dma_irq_flag==1 && rc==0 is normal completion
     * rc==-ERESTARTSYS is user abort
     * other cases not to happen, but are treated as -ECANCELED
     */
    cond1 = board->dma1_irq_flag;
    board->dma1_irq_flag = 0;
	board->read1_in_progress = 0;
    dev_dbg(&board->pci_dev->dev, "read() DMA1 wait complete w/ rc=%d cond1=%d\n", rc, cond1);

    // TODO: not understood
	if ((rc != 0) || (cond1 != 1 )) 
    { /* interrupted or aborted */
		if(cond1!=1) rc = -ECANCELED;
		/* reset DMA engine */
		dma1_reset(board);
        board->dma1_bytes_trans = 0;
        spin_unlock_irq(&board->dma_queue.lock);

        dev_warn(&board->pci_dev->dev, "DMA1 read(): interrupt failed: %d\n", rc);
		return rc;
	} 
    else 
    {
        // If everithing went smooth till now
        spin_unlock_irq(&board->dma_queue.lock);

		i = 0;
        dev_dbg(&board->pci_dev->dev, "  read(): DMA1 returned from sleep %lu\n", dma1_count);
		rc = 0;
        // copy data from RAM to User memory (in the case when tmp_count > DMA_BUF_SIZE)
		while ((dma1_count > DMA_BUF_SIZE) && (rc == 0)) 
        {
			rc = copy_to_user(buf + DMA_BUF_SIZE*i, board->kernel_mem1_buf[i], DMA_BUF_SIZE);
            dev_dbg(&board->pci_dev->dev, "  copy_to_user() DMA1: n bytes not copied %d\n", rc);
			dma1_count -= DMA_BUF_SIZE;
			/* sometimes the DMA done interrupt comes even though nothing has been
			 * transfered.  Fill our buffer with a test pattern so that this is more
			 * obvious.
			 */
			memset(board->kernel_mem1_buf[i], 0xf0, DMA_BUF_SIZE);
			i++;
		}
        // copy DMA1 data from RAM to User memory (in the case when tmp_count < DMA_BUF_SIZE)
		if (rc == 0) 
        {
            // copy_to_user return the number of bytes that could NOT be copied
			rc = copy_to_user(buf + DMA_BUF_SIZE*i,	board->kernel_mem1_buf[i], dma1_count);
			memset(board->kernel_mem1_buf[i], 0xf0, dma1_count);
		}

		if (rc) 
        {
            return rc;
        }

	}

    //////////////////////////////////////////////////////////
    // check DMA2
    //////////////////////////////////////////////////////////
    spin_lock_irq(&board->dma_queue.lock);
    if (likely(board->irqmode != dmac_irq_poll)) 
    {
        //const unsigned long twait2 = msecs_to_jiffies(2);
        // TODO: For the moment I check only the dma1_queue, check also dma2_queue??
        // wait_event_interruptible_locked_irq - sleep until a condition gets true 
        // The function will return -ERESTARTSYS if it was interrupted by a signal and 0 if condition evaluated to true.
        dev_dbg(&board->pci_dev->dev, "  start wait_event_interruptible_locked_irq on DMA2\n");
        rc = wait_event_interruptible_locked_irq(board->dma_queue, board->dma2_irq_flag!=0);
        dev_dbg(&board->pci_dev->dev, "  wait_event_interruptible_locked_irq on DMA2 rc = %d\n", rc);
    } 
    else 
    {
        const unsigned long twait1 = msecs_to_jiffies(1);
        do 
        {
            spin_unlock_irq(&board->dma_queue.lock);
            /* must unlock for call to amc_isr() as spin locks aren't recursive */
            if (amc_isr(board->pci_dev->irq, board) == IRQ_NONE)
            {
                // wait_event_interruptible_timeout - sleep until a condition gets true or a timeout elapses
                // 0 if the condition evaluated to false after the timeout elapsed, 1 if the condition evaluated 
                // to true after the timeout elapsed, the remaining jiffies (at least 1) if the condition evaluated 
                // to true before the timeout elapsed, or -ERESTARTSYS if it was interrupted by a signal
                rc = wait_event_interruptible_timeout(board->dma_queue, board->dma2_irq_flag!=0, twait1);
                dev_warn(&board->pci_dev->dev, "  wait_event_interruptible_timeout on DMA2 rc = %d\n", rc);
            }
            else
            {
                rc = 0;
            }
            spin_lock_irq(&board->dma_queue.lock);
            /* continue while no "IRQ" signaled, and wait not interrupted */
        } while(board->dma2_irq_flag==0 && rc>=0);

        if (rc > 0)
        {
            rc = 0;
        }

    }
    /*
     * dma_irq_flag==1 && rc==0 is normal completion
     * rc==-ERESTARTSYS is user abort
     * other cases not to happen, but are treated as -ECANCELED
     */

    cond2 = board->dma2_irq_flag;
    board->dma2_irq_flag = 0;
	board->read2_in_progress = 0;
    dev_dbg(&board->pci_dev->dev, "read() DMA2 wait complete w/ rc=%d cond2=%d\n", rc, cond2);

    // TODO: not understood
	if ((rc != 0) || (cond2 != 1 )) 
    { /* interrupted or aborted */
		if(cond2!=1) rc = -ECANCELED;
		/* reset DMA engine */
		dma2_reset(board);
        board->dma2_bytes_trans = 0;
        spin_unlock_irq(&board->dma_queue.lock);

        dev_warn(&board->pci_dev->dev, "DMA2 read(): interrupt failed: %d\n", rc);
		return rc;
	} 
    else 
    {
        // If everithing went smooth till now
        spin_unlock_irq(&board->dma_queue.lock);

        dev_dbg(&board->pci_dev->dev, "  read(): DMA2 returned from sleep %lu\n", dma2_count);
        dev_info(&board->pci_dev->dev, "  read(): DMA2 returned from sleep %lu, %lu, %d\n", dma2_count, DMA_BUF_SIZE, i);
		rc = 0;
        // copy data from RAM to User memory (in the case when tmp_count > DMA_BUF_SIZE)
		rc = copy_to_user(buf + DMA_BUF_SIZE*i + dma1_count, board->kernel_mem2_buf, dma2_count);
		memset(board->kernel_mem2_buf, 0xf0, dma2_count);
		
		if (rc) 
        {
            return rc;
        }

	}

	*pos += count;

	return count;
}


/* all possible ioctl() value types */
union ioctl_value {
    uint8_t u8;
    uint32_t u32;
    struct trg_ctrl trg;
    struct pico_float_coeff float_coeff;
};

static
long char_ioctl(
	struct file *filp,
	unsigned int cmd,
	unsigned long arg
)
{
    union ioctl_value uval;
	long ret;
    struct file_data *fdata = (struct file_data *)filp->private_data;
    struct board_data *board = fdata->board;
    size_t tocpy = sizeof(uval);

    /* copy some bytes to/from user space, others are zero'd */
    memset(&uval, 0, sizeof(uval));
    if(_IOC_SIZE(cmd)<sizeof(uval))
        tocpy = _IOC_SIZE(cmd);

    dev_info(&board->pci_dev->dev, "%s: 0x%08x size=%u\n", __PRETTY_FUNCTION__, cmd, (unsigned)tocpy);

    if(_IOC_DIR(cmd)&_IOC_WRITE) {
        /* copy in all provided bytes. based on IOCTL code. */
        ret = copy_from_user(&uval, (const void*)arg, tocpy);
        if(ret) return ret;
    }

	/* validate cmd and copy in values from user before locking
	 * can't access board->
	 */
	switch (cmd) {
    case SET_RANGE:
        ret = 0;
		break;
    case SET_FSAMP:	{
		uint32_t scaler = PICO_CLK_FREQ / uval.u32 - 1;
		if ((uval.u32 > PICO_ADC_MAX_FREQ) || (scaler > (PICO_CONV_MAX-1))){
			return -EINVAL;
		} else {
			ret = 0;
		}
		break;
	}
    case SET_TRG:
    case SET_RING_BUF:
	case SET_GATE_MUX:
	case SET_CONV_MUX:
	case GET_RANGE:
	case GET_FSAMP:
	case GET_B_TRANS:
    case GET_THRESHOLD_FLAG:
    case GET_THRESHOLD:
    case SET_THRESHOLD:
    case GET_NSAMP:
    case SET_NSAMP:
    case GET_MPS_CONFIG:
    case SET_MPS_CONFIG:
    case SET_MPS_CTRL:
    case GET_TRG_ACC_DATA:   // read accumulator output data in Trigger Mode
    case GET_TRG_CNTR_1M:       // read accumulator counters in Trigger Mode
    case GET_TRG_CNTR_1K:
    case SET_TRG_NRSAMP:
    case GET_TRG_NRSAMP:
    case SET_TRG_DELAY:
    case GET_TRG_DELAY:
	case ABORT_READ:
		ret = 0;
		break;
	case GET_VERSION:
        /* Versions:
         *  0 - implied by errno==EINVAL
         *  1 - Added GET_VERSION and ABORT_READ
         *  2 - Added GET_SITE_ID, GET_SITE_VERSION, SET_SITE_MODE.
         *      Changed all others.
         *  3 - Changed GET_FSAMP and SET_FSAMP to use frequency as
         *      a parameter
         */
        return put_user(GET_VERSION_CURRENT, (uint32_t*)arg);
    case GET_SITE_ID:
        return put_user(dmac_site, (uint32_t*)arg);
    case GET_SITE_VERSION:
    {
        uint32_t sver;
        switch(dmac_site) {
#ifdef CONFIG_MPS_PICO_FRIB
        case USER_SITE_FRIB: sver = 0; break;
#endif
        default: sver = 0;
        }
        return put_user(sver, (uint32_t*)arg);
    }
    case SET_SITE_MODE:
        if(0) {}
#ifdef CONFIG_MPS_PICO_FRIB
        else if(dmac_site==USER_SITE_FRIB) {
            if(uval.u32>2) return -EINVAL;
        }
#endif
        else if(uval.u32!=0) return -EINVAL;

        fdata->site_mode = uval.u32;

        return 0;
	default:
        ret = -EINVAL;
	}

    if(ret) return ret;

    /* locking here to protect RMW register operations.
     * Use dma_queue.lock for convinience
     */
    spin_lock_irq(&board->dma_queue.lock); /* enter critical section, can't sleep */
    // Uncomment these two lines to use DMA1 or DMA2 singularly
    // spin_lock_irq(&board->dma1_queue.lock); /* enter critical section, can't sleep */
    // spin_lock_irq(&board->dma2_queue.lock); /* enter critical section, can't sleep */

	switch (cmd) {
    case SET_RANGE: {
        uint32_t control;
		control = ioread32(board->bar0);
		control &= ~0xFFUL;
        control |= uval.u8;
		iowrite32(control, board->bar0);
		control = ioread32(board->bar0);
		break;
    }
	case GET_RANGE:
        uval.u8 = ioread32(board->bar0) & 0xFF;
		break;

	case SET_FSAMP: {
        uint32_t scaler = PICO_CLK_FREQ / uval.u32 - 1;
        iowrite32(scaler, board->bar0 + PICO_CONV_GEN);
        uval.u32 = ioread32(board->bar0 + PICO_CONV_GEN);
		break;
	}

	case GET_FSAMP: {
        uint32_t scaler = ioread32(board->bar0 + PICO_CONV_GEN);
        uval.u32 = PICO_CLK_FREQ / (scaler + 1);
		break;
	}

	case GET_B_TRANS: {
        uval.u32 = board->dma1_bytes_trans + board->dma2_bytes_trans; // - DMA_BYTES_PER_LINE;  // TODO:
		break;
    }

    case SET_TRG: {
        uint32_t ctrl_tmp;
        iowrite32(*(uint32_t *)&uval.trg.limit,
			board->bar0 + PICO_ADDR + TRG_OFFS_LIMIT);
        iowrite32(uval.trg.nr_samp,
			board->bar0 + PICO_ADDR + TRG_OFFS_NRSAMP);

		ctrl_tmp = ioread32(board->bar0 + PICO_ADDR + TRG_OFFS_CTRL);

		/* change the trigger edge */
		ctrl_tmp &= ~(0x3);
        ctrl_tmp |= uval.trg.mode;

		/* change the channel bits */
		ctrl_tmp &= ~(0x7 << TRG_CTRL_CH_SHIFT);
        ctrl_tmp |= uval.trg.ch_sel << TRG_CTRL_CH_SHIFT;

		iowrite32(ctrl_tmp, board->bar0 + PICO_ADDR + TRG_OFFS_CTRL);
        uval.u32 = ioread32(board->bar0 + PICO_ADDR + TRG_OFFS_CTRL);
		break;
    }
	case SET_RING_BUF:
        iowrite32(uval.u32,
			board->bar0 + PICO_ADDR + RING_BUFF_OFFS_DELAY);
		break;

    case SET_GATE_MUX: {
        uint32_t ctrl_tmp;
        uval.u32 &= MUX_TRG_MASK;
        uval.u32 <<= MUX_TRG_SHIFT;

		ctrl_tmp = ioread32(board->bar0 + PICO_ADDR + PICO_CONV_TRG);
		ctrl_tmp &= ~(MUX_TRG_MASK << MUX_TRG_SHIFT);
        ctrl_tmp |= uval.u32;

		iowrite32(ctrl_tmp, board->bar0 + PICO_ADDR + PICO_CONV_TRG);
        uval.u32 = ioread32(board->bar0 + PICO_ADDR + PICO_CONV_TRG);
		break;
    }
    case SET_CONV_MUX: {
        uint32_t ctrl_tmp;
        uval.u32 &= MUX_CONV_MASK;
        uval.u32 <<= MUX_CONV_SHIFT;

		ctrl_tmp = ioread32(board->bar0 + PICO_ADDR + PICO_CONV_TRG);
		ctrl_tmp &= ~(MUX_CONV_MASK << MUX_CONV_SHIFT);
        ctrl_tmp |= uval.u32;

		iowrite32(ctrl_tmp, board->bar0 + PICO_ADDR + PICO_CONV_TRG);
        uval.u32 = ioread32(board->bar0 + PICO_ADDR + PICO_CONV_TRG);
		break;
    }
    case GET_THRESHOLD_FLAG: {
        uint32_t address = uval.float_coeff.addr * (0x04);
        uint32_t *ptr = (uint32_t *)&uval.float_coeff.data;
        *ptr = ioread32(board->bar0 + PICO_MOVING_AVERAGE + OVER_THRESHOLD_BA + address);
        break;
    }
    case GET_THRESHOLD: {
        uint32_t address = uval.float_coeff.addr * (0x04);
        uint32_t *ptr = (uint32_t *)&uval.float_coeff.data;
        *ptr = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_THRESHOLD + address);
        break;
    }
    case SET_THRESHOLD: {
        uint32_t address = uval.float_coeff.addr * (0x04);
        iowrite32(*(uint32_t *)&uval.float_coeff.data, board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_THRESHOLD + address);
        break;
    }
    case GET_NSAMP: {
        uint32_t address = uval.float_coeff.addr * (0x04);
        uint32_t *ptr = (uint32_t *)&uval.float_coeff.data;
        *ptr = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_NSAMP + address);
        break;
    }
    case SET_NSAMP: {
        uint32_t address = uval.float_coeff.addr * (0x04);
        iowrite32(*(uint32_t *)&uval.float_coeff.data, board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_NSAMP + address);
        break;
    }
    case GET_MPS_CONFIG: {
        uval.u32 = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_MPS_CONFIG);
        break;
    }
    case SET_MPS_CONFIG: {
        iowrite32(uval.u32, board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_MPS_CONFIG);
        break;
    }
    case SET_MPS_CTRL: {
        iowrite32(uval.u32, board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_CTRL);
        break;
    }
    case GET_TRG_ACC_DATA: {
        uint32_t address = uval.float_coeff.addr * (0x04);
        uint32_t *ptr = (uint32_t *)&uval.float_coeff.data;
        *ptr = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_ACC_1M + address);
        break;
    }
    case GET_TRG_CNTR_1M: {
        uval.u32 = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_CNTR_1M);
        break;
    }
    case GET_TRG_CNTR_1K: {
        uval.u32 = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_CNTR_1K);
        break;
    }
    case SET_TRG_NRSAMP: {
        iowrite32(uval.u32, board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_NRSAMP);
        break;
    }
    case GET_TRG_NRSAMP: {
        uval.u32 = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_NRSAMP);
        break;
    }
    case SET_TRG_DELAY: {
        iowrite32(uval.u32, board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_DELAY);
        break;
    }
    case GET_TRG_DELAY: {
        uval.u32 = ioread32(board->bar0 + PICO_MOVING_AVERAGE + PICO_MA_TRG_DELAY);
        break;
    }
	case ABORT_READ:
        /* abort in progress DMA waiter */ 
        board->dma1_irq_flag = 2;
        board->dma2_irq_flag = 2;
        wake_up_locked(&board->dma_queue);
        // Uncomment these two lines if you want to use DMA1 or DMA2 singularly
        // wake_up_locked(&board->dma1_queue);
        // wake_up_locked(&board->dma2_queue);

	default:
        dev_info(&board->pci_dev->dev, "%s:   unknown ioctl\n", __PRETTY_FUNCTION__);
		break;
	}

    spin_unlock_irq(&board->dma_queue.lock);
    // Uncomment these two lines if you want to use DMA1 or DMA2 singularly
    // spin_unlock_irq(&board->dma1_queue.lock); /* end of critical section, can't access board-> */
    // spin_unlock_irq(&board->dma2_queue.lock); /* end of critical section, can't access board-> */

#ifdef CONFIG_MPS_PICO_FRIB
    if(cmd==ABORT_READ) {
        /* abort any waiting for capture buffer */
        spin_lock_irq(&board->capture_queue.lock);
        board->capture_ready = 2;
        wake_up_locked(&board->capture_queue);
        spin_unlock_irq(&board->capture_queue.lock);
    }
#endif

    if(_IOC_DIR(cmd)&_IOC_READ) {
        ret = copy_to_user((void*)arg, &uval, tocpy);
        if(ret) return ret;
    }

	return ret;
}

static
ssize_t char_write(struct file *filp, const char __user *buf, size_t count, loff_t *pos)
{
    struct file_data *fdata = (struct file_data *)filp->private_data;
    struct board_data *board = fdata->board;
    (void)board;
    if(0) {}
#ifdef CONFIG_MPS_PICO_FRIB
    else if(dmac_site==USER_SITE_FRIB) {
        switch(fdata->site_mode) {
        case 1:  return frib_write_reg(board, buf, count, pos);
        default: return -EINVAL;
        }
    }
#endif
    else
        return -EINVAL;
}

static
loff_t char_llseek(struct file *filp, loff_t pos, int whence)
{
    struct file_data *fdata = (struct file_data *)filp->private_data;
    //struct board_data *board = fdata->board;
    loff_t npos;

    if(dmac_site!=USER_SITE_FRIB || fdata->site_mode==0)
        return -EINVAL;
#ifdef CONFIG_MPS_PICO_FRIB

    switch(whence) {
    case 0: npos = pos; break;
    case 1: npos = filp->f_pos + pos; break;
    case 2: npos = 0x40000 + pos; break;
    default: return -EINVAL;
    }

    if(npos<0)
        return -EINVAL;
    else if(npos>0x40000)
        npos = 0x40000;

    filp->f_pos = npos;
    return npos;

#else
    //(void)board;
    (void)npos;
    return -EINVAL;
#endif
}

const struct file_operations mps_pico_fops = {
	.owner		= THIS_MODULE,
	.open		= char_open,
	.release	= char_release,
	.read		= char_read,
    .write      = char_write,
    .llseek     = char_llseek,
	.unlocked_ioctl = char_ioctl
};

#ifdef CONFIG_MPS_PICO_FRIB

static ssize_t frib_write_reg (
    struct board_data *board,
    const char __user *buf,
    size_t count,
    loff_t *pos)
{
    ssize_t ret=0;
    unsigned offset, end;
    const uint32_t __user *ibuf = (const uint32_t __user *)buf;

    if(count%4) return -EINVAL;
    if(!pos || *pos<USER_ADDR) return -EINVAL;

    offset = *pos-USER_ADDR;

    if(offset>=0x10000) return -EINVAL;

    if(count>0x10000-offset)
        count = 0x10000-offset;

    end = offset+count;

    for(; offset<count; offset+=4, ibuf+=4)
    {
        uint32_t val;
        ret = get_user(val, ibuf);
        if(unlikely(ret)) break;
        iowrite32(val, board->bar0 + USER_ADDR + offset);
    }

    if(!ret) ret=count;
    return ret;
}

static ssize_t frib_read_reg (
    struct board_data *board,
    char __user *buf,
    size_t count,
    loff_t *pos)
{
    ssize_t ret=0;
    unsigned offset, end;
    uint32_t __user *ibuf = (uint32_t __user *)buf;

    if(count%4) return -EINVAL;
    if(!pos || *pos<USER_ADDR) return -EINVAL;

    offset = *pos-USER_ADDR;

    if(offset>=0x10000) return 0;

    if(count>0x10000-offset)
        count = 0x10000-offset;

    end = offset+count;

    for(; ret && offset<count; offset+=4, ibuf+=4)
    {
        uint32_t val = ioread32(board->bar0 + USER_ADDR + offset);
        ret = put_user(val, ibuf);
    }

    if(!ret) ret=count;
    return ret;
    /* *pos not updated */
}

static ssize_t frib_read_capture (
    struct board_data *board,
    char __user *buf,
    size_t count,
    loff_t *pos)
{
    ssize_t ret;
    unsigned offset;

    if(count%4) return -EINVAL;
    if(!pos || *pos<USER_ADDR+0x100) return -EINVAL;

    offset = *pos-(USER_ADDR+0x100);

    if(offset>=board->capture_length) {
        dev_info(&board->pci_dev->dev, "Capture read starts past end of buffer %u %u\n",
                offset, board->capture_length);
        return 0;
    }

    if(count > board->capture_length-offset)
        count = board->capture_length-offset;

    spin_lock_irq(&board->capture_queue.lock);

    ret = wait_event_interruptible_locked_irq(board->capture_queue, board->capture_ready!=0);

    if(!ret && board->capture_ready!=1)
        ret = -ECANCELED;
    board->capture_ready = 0;

    spin_unlock_irq(&board->capture_queue.lock);

    if(ret) return ret;

    ret = copy_to_user(buf, board->capture_buf, count);
    if(!ret) ret=count;
    return ret;
    /* *pos not updated */
}

#endif
