
// Copyright (c) 2015 CAEN ELS d.o.o.

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <malloc.h>

#define BYTES_PER_LINE	(64)    // this must be a multiple of 32 otherwise it is not compatible with LINUX PAGE SIZE (usually 4kByte)  
#define OUT_FORMAT		"%+0.6e"
#define OUT_FORMAT_CNTR "%u"

char MPS_PICO_DEV[128];


////////////////////////////////////////////////////////////////////////////////
/// \brief prints usage information

void print_usage(const char* name){
	printf("AMC-Pico-8 measurement dump\n");
	printf("\n");
	printf("Arguments:\n");
	printf("    --devfile DEVFILE  Device file in /dev");
	printf("    --nrsamp NRSAMP  Number of samples to read\n");
	printf("    --out FILENAME   Output file name\n");
	printf("    --force          Overwrites output file\n");
	printf("\n");
	printf("Example:\n");
	printf("    %s --devfile /dev/mps_pico_0000:03:00.0 --nrsamp 100 --out meas1.csv", name);
	printf("\n");
}

////////////////////////////////////////////////////////////////////////////////
/// \brief reads from AMC-Pico-8

int read_from_pico(uint32_t nrsamp, void *buf) {

	int fd = open(MPS_PICO_DEV, O_RDONLY | O_SYNC);
	if (fd<0) {
		perror("open()");
		return -1;
	}

	ssize_t bytes_read = read(fd, buf, nrsamp*BYTES_PER_LINE);

	close(fd);
	return !(bytes_read == nrsamp*BYTES_PER_LINE);
}

////////////////////////////////////////////////////////////////////////////////
/// \brief dumps buffer to CSV file

int dump_to_file(void* buf, const char* filename, uint32_t nrsamp, int force) {

	FILE* f = fopen(filename, force ? "w" : "wx");
	if (f == NULL) {
		perror("fopen()");
		return -1;
	}

	// Print header
	fprintf(f, "Index,");
	for (int ch=0; ch<8; ch++)
		fprintf(f, "Channel %d,", ch);
	fprintf(f, "\n");

	float* d_ptr = (float*) buf;
	uint32_t* cntr_ptr;
	uint32_t cntr;
	uint32_t n_buffer;
	for (uint32_t line=0; line<nrsamp; line++){
		// print line number
		fprintf(f, "%d,", line);
		// print counter
		// cntr_ptr = (uint32_t*)d_ptr;
		// fprintf(f, OUT_FORMAT_CNTR ",", *cntr_ptr);
		// d_ptr++;
		
		//n_buffer = cntr >> 24;
		//cntr = cntr & 0x00FFFFFF;
		
		// print channel data
		for (int ch=0; ch<8; ch++)
			fprintf(f, OUT_FORMAT ",", *d_ptr++);

		cntr_ptr = (uint32_t*) d_ptr;
		cntr = (*cntr_ptr) & 0x00FFFFFF;
		n_buffer = ((*cntr_ptr) >> 24) & 0xFF;
		d_ptr += 8;
		fprintf(f, "%u,%u,", n_buffer, cntr);
		fprintf(f, "\n");
	}

	fclose(f);

	return 0;
}


////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv) {

	uint32_t nrsamp = 0;
	char filename[256] = {0};
	int force = 0;

	struct timeval st, et;

	static struct option long_options[] = {
		{"help",    no_argument,       NULL, 'h' },
		{"devfile", required_argument, NULL, 'd' },
		{"nrsamp",  required_argument, NULL, 'n' },
		{"out",     required_argument, NULL, 'o' },
		{"force",   no_argument, NULL, 'f' },
		{0, 0, 0, 0 }
	};

	while (1) {
		int c;
		c = getopt_long(argc, argv, "n:o:fh", long_options, NULL);
		if (c == -1)
			break;

		switch(c){
		case 'h':
			print_usage(argv[0]);
			return 0;
		case 'd':
			strncpy(MPS_PICO_DEV, optarg, sizeof(MPS_PICO_DEV));
		case 'n':
			sscanf(optarg, "%i", &nrsamp);
			break;
		case 'o':
			strncpy(filename, optarg, 255);
			break;
		case 'f':
			force = 1;
			break;
		}
	}

	if ((nrsamp != 0) && strlen(filename)) {
		size_t tot_bytes = BYTES_PER_LINE*nrsamp;
		void *buf = malloc(tot_bytes);
		if (buf == NULL) {
			perror("malloc()");
			return -1;
		}
		size_t allocated_bytes = malloc_usable_size(buf);
		printf("Byte requested: %lu, Allocated: %lu\n", tot_bytes, allocated_bytes);

		int rc;
		gettimeofday(&st,NULL); 
		rc = read_from_pico(nrsamp, buf);
		gettimeofday(&et,NULL); 
		if (rc < 0){
			printf("Error reading from AMC-Pico-8\n");
			free(buf);
			return -1;
		}		

		int elapsed = ((et.tv_sec - st.tv_sec) * 1000000) + (et.tv_usec - st.tv_usec);
		printf("Read time: %d micro seconds\n", elapsed);

		dump_to_file(buf, filename, nrsamp, force);

		free(buf);
	}

	return 0;
}
